/*
	KUMAamp project()
	Copyright (C) Mamiya 2000.
	License:LGPL
	---------------------------------------------------------------------------
	PMD sequencer(PMDWin interface)
	seq_pmd.cpp
*/

#include "in_pmd.h"

/* ANSI/Windows standard headers */
#include <windows.h>
#include <shlobj.h>

/* Libraries headers */
#include <pmdwin.h>

/* Project headers */
#include "../sequencer.h"
#include "in_pmd.rh"

#define MAX_CH 32

/* PMD sequencer struct */
typedef struct {
	SEQUENCER seq;			/* KUMAamp common interface */
	int loopc;				/* zero based loopcounter */
	int loopcount;			/* max loop count */
	int frequency;			/* mixing frequency */
	int fade;				/* 0:normal 1-:fading out(dB/(1 >> FADE_SHIFT)) */
	int fadetime;			/* fadeout time(ms) */

	int volFM;				/* unknown(0-255) */
	int volSSG;				/* unknown(0-255) */
	int volPPZ;				/* unknown(0-255) */
	int volADPCM;			/* unknown(0-255) */
	int volRHYTHM;			/* unknown(0-255) */

	int fPPS;				/* Enable PPS */
	int fRHYTHM;			/* RHYTHM with PPS */
	int fPMDB2;				/* Compatible PMDB2 PCM */

	int chenable[MAX_CH];

} PMDSEQ;

/* PMDamp setting struct */
static struct {
	/* OPNA rythm wav files path */
	char rhythmpath[MAX_PATH];
	char currentrhythmpath[MAX_PATH];
	int frequency;			/* mixing frequency */
	int priority;			/* play thread priority */
	int loopcount;			/* max loop count */
	int fadetime;			/* fadeout time(ms) */

	int volFM;				/* unknown(0-255) */
	int volSSG;				/* unknown(0-255) */
	int volPPZ;				/* unknown(0-255) */
	int volADPCM;			/* unknown(0-255) */
	int volRHYTHM;			/* unknown(0-255) */

	int fPPS;				/* Enable PPS */
	int fRHYTHM;			/* RHYTHM with PPS */
	int fPMDB2;				/* Compatible PMDB2 PCM */

	int chenable[MAX_CH];

	HINSTANCE hDllInstance;
	char cfgname[MAX_PATH];

	/* PCM search path list */
	int pcmdirnum;
	char *(pcmdirlist[65]);
	char pcmdir[64][MAX_PATH];
} setting;

typedef struct {
	char *key;
	int num;
} KEYINT;

static KEYINT freqlist[] = {
	{ "55467:FM",	SOUND_55K },
	{ "44100",		SOUND_44K },
	{ "22050",		SOUND_22K },
	{ "11025",		SOUND_11K },
	{ NULL,			SOUND_55K },
};

static KEYINT priolist[] = {
	{ "Highest",	5 },
	{ "Higher",		4 },
	{ "Normal",		3 },
	{ "Lower",		2 },
	{ "Lowest",		1 },
	{ NULL,			3 },
};

#define PTR__ ((PMDSEQ *)(this__->work))
static void Play(SEQUENCER *this__)
{
	int i;
	for (i = 0; i < MAX_CH; i++)
		PTR__->chenable[i] = -1;
	PTR__->volFM = PTR__->volSSG = PTR__->volPPZ = -1;
	PTR__->volADPCM = PTR__->volRHYTHM = -1;
	PTR__->fPPS = PTR__->fRHYTHM = PTR__->fPMDB2 -1;

	if (lstrcmpi(setting.currentrhythmpath, setting.rhythmpath))
	{
		loadrhythmsample(setting.rhythmpath);
		lstrcpy(setting.currentrhythmpath, setting.rhythmpath);
	}
	PTR__->loopc = 0;
	PTR__->loopcount = setting.loopcount;
	PTR__->fadetime = setting.fadetime;
	PTR__->fade = 0;
	setpcmrate(PTR__->frequency);

	setppsuse(TRUE);
	music_start();
}
static void Stop(SEQUENCER *this__)
{
	PTR__->loopc = 0;
	music_stop();
}

static void Mix(SEQUENCER *this__, void *buf, int numsamp)
{
	int i;
	for (i = 0; i < MAX_CH; i++)
		if (PTR__->chenable[i] != setting.chenable[i])
		{
			PTR__->chenable[i] = setting.chenable[i];
			if (PTR__->chenable[i])
				maskoff(i);
			else
				maskon(i);
		}
#define SETTING_XY(x,y) if (PTR__->y != setting.y) x(PTR__->y = setting.y);
	SETTING_XY(setfmvoldown, volFM);
	SETTING_XY(setssgvoldown, volSSG);
	SETTING_XY(setppzvoldown, volPPZ);
	SETTING_XY(setadpcmvoldown, volADPCM);
	SETTING_XY(setrhythmvoldown, volRHYTHM);
#undef SETTING_XY
#define SETTING_XY(x,y) if (PTR__->y != setting.y) x((PTR__->y = setting.y) != 0);
	SETTING_XY(setppsuse, fPPS);
	SETTING_XY(setrhythmwithssgeffect, fRHYTHM);
	SETTING_XY(setpmd86pcmmode, fPMDB2);
#undef SETTING_XY

	getpcmdata((short *)buf, numsamp);

	OPEN_WORK *openworkp = getopenwork();
	PTR__->loopc = (openworkp->status2 > 0) ? openworkp->status2 : 0;
	if (PTR__->loopcount == 0 || PTR__->fade != 0) return;
	if (openworkp->status2 < PTR__->loopcount) return;
	PTR__->fade = 1;
	openworkp->fade_stop_flag = 1;
	if (PTR__->fadetime != 0)
		fadeout2(PTR__->fadetime);
	else
		Stop(this__);
}
static void Term(SEQUENCER *this__)
{
	Stop(this__);
	delete PTR__;
}
static void SetPosition(SEQUENCER *this__, int time_in_ms)
{
	setpos(time_in_ms);
}
static int GetPosition(SEQUENCER *this__)
{
	return getpos();
}
static int GetRate(SEQUENCER *this__)
{
	PTR__->frequency = setting.frequency;
	return (PTR__->frequency == SOUND_55K) ? SOUND_44K : PTR__->frequency;
}
static int GetPriority(SEQUENCER *this__)
{
	return setting.priority;
}

static int GetLoopCount(SEQUENCER *this__)
{
	return PTR__->loopc;
}

static int IsPlaying(SEQUENCER *this__)
{
	OPEN_WORK *openworkp = getopenwork();
	return (openworkp->play_flag != 0) && (openworkp->status2 != -1);
}

#ifdef SINGLE_BUILD
#define InitPmdSequencer InitSequencer
#define QuitPmdSequencer QuitSequencer
#define AboutPmdSequencer AboutSequencer
#define ConfigPmdSequencer ConfigSequencer
#define getPmdFileInfo getFileInfo
#define loadPmdFile loadFile
#define infoPmdBox infoBox
#define isOurPmdFile isOurFile
#endif

static void WritePrivateProfileInt(LPCSTR sect, LPCSTR key, UINT data, LPCSTR file)
{
	char temp[32], *p;
	temp[30] = '0';
	temp[31] = '\0';
	if (data == 0)
		p = &temp[30];
	else
		for (p = &temp[31]; data != 0; data /= 10)
		{
			*(--p) = '0' + (data % 10);
		}
	WritePrivateProfileString(sect, key, p, file);
}

static void GetSettingString(LPCSTR lpKey, LPCSTR lpDefault, LPSTR lpBuf, DWORD nSize)
{
	GetPrivateProfileString("PMDWIN", lpKey, lpDefault, lpBuf, nSize, setting.cfgname);
}
static void SetSettingString(LPCSTR lpKey, LPCSTR lpStr)
{
	WritePrivateProfileString("PMDWIN", lpKey, lpStr, setting.cfgname);
}
static int GetSettingInt(LPCSTR lpKey, int iDefault)
{
	char buf[128], dbuf[128];
	wsprintf(dbuf, "%d", iDefault);
	GetPrivateProfileString("PMDWIN", lpKey, dbuf, buf, sizeof(buf), setting.cfgname);
	return atoi(buf);
}
static void SetSettingInt(LPCSTR lpKey, int iInt)
{
	char buf[128];
	wsprintf(buf, "%d", iInt);
	WritePrivateProfileString("PMDWIN", lpKey, buf, setting.cfgname);
}

#include "cfpmdwin.h"
#include "cfpmdmix.h"

void InitPmdSequencer(HINSTANCE hDllInstance)
{
	char path[MAX_PATH];
	LPTSTR fnoff;

	setting.hDllInstance = hDllInstance;

	GetModuleFileName(hDllInstance, path, MAX_PATH);
	GetFullPathName(path, MAX_PATH, setting.cfgname, &fnoff);
	setting.cfgname[(fnoff - setting.cfgname)] = '\0';
	lstrcat(setting.cfgname, "plugin.ini");

	PMDWINLoadSetting();
	PMDMIXLoadSetting();
}

void QuitPmdSequencer(HINSTANCE hDllInstance)
{
}

void AboutPmdSequencer(HWND hwndParent)
{
	MessageBox(hwndParent, "PMD for Windows uPMDWinv Ver.0.13\n\tCopyright & Programmed by C60\n\nM88 - PC-88 Emulator | FM Sound Generator - Core Unit\n\tCopyright (C) cisc 1998, 1999.\n\nPMDamp\n\t(C) 2000 Mamiya\n\t\tmailto:mamiya@proc.org.tohoku.ac.jp\n\t\thttp://www.proc.org.tohoku.ac.jp/befis/", PLUGIN_NAME, MB_OK);
}

void ConfigPmdSequencer(HWND hwndParent)
{
	PROPSHEETPAGE psp[2];
	PROPSHEETHEADER psh;

	ZeroMemory(psp, sizeof(psp));
	psp[0].dwSize = sizeof(PROPSHEETPAGE);
	psp[0].dwFlags = PSP_DEFAULT;
	psp[0].hInstance = setting.hDllInstance;
	psp[0].pszTemplate = MAKEINTRESOURCE(IDD_PSP_PMDWIN);
	psp[0].pszIcon = NULL;
	psp[0].pszTitle = NULL;
	psp[0].pfnDlgProc = PMDWINSettingProc;
	psp[0].lParam = 0;
	psp[0].pfnCallback = NULL;
	psp[0].pcRefParent = NULL;

	psp[1].dwSize = sizeof(PROPSHEETPAGE);
	psp[1].dwFlags = PSP_DEFAULT;
	psp[1].hInstance = setting.hDllInstance;
	psp[1].pszTemplate = MAKEINTRESOURCE(IDD_PSP_PMDMIX);
	psp[1].pszIcon = NULL;
	psp[1].pszTitle = NULL;
	psp[1].pfnDlgProc = PMDMIXSettingProc;
	psp[1].lParam = 0;
	psp[1].pfnCallback = NULL;
	psp[1].pcRefParent = NULL;

	ZeroMemory(&psh, sizeof(psh));
	psh.dwSize = sizeof(PROPSHEETHEADER);
	psh.dwFlags = PSH_NOAPPLYNOW | PSH_PROPSHEETPAGE;
	psh.hwndParent = hwndParent;
	psh.hInstance = setting.hDllInstance;
	psh.pszIcon = NULL;
	psh.pszCaption = (LPSTR) "PMDamp setting";
	psh.nPages = sizeof(psp) / sizeof(PROPSHEETPAGE);
	psh.nStartPage = 0;
	psh.ppsp = (LPCPROPSHEETPAGE) &psp;
	psh.pfnCallback = NULL;
	PropertySheet(&psh);
}

int getPmdFileInfo(char *fn, char *title, int *length_in_ms)
{
	extern PMDWIN pmdwin2;
	if (length_in_ms != NULL)
	{
		int length = 0, loop = 0;
		if (false == pmdwin2.getlength(fn, &length, &loop)) return false;
		*length_in_ms = length/* + loop */;
	}
	else if (title != NULL)
	{
		int result;
		result = pmdwin2.music_load(fn);
		if(result == ERR_OPEN_MUSIC_FILE || result == ERR_WRONG_MUSIC_FILE || result == ERR_OUT_OF_MEMORY) return false;
	}

	if (title != NULL)
	{
		char *p, *p2;
		title[0] = '\0';
		pmdwin2.getmemo2(title,NULL,0,1);
		for (p2 = p = fn;1;)
		{
			char *prev;
			if (*p == '/' || *p == '\\' || *p == ':') p2 = p + 1;
			p = CharNext(prev = p);
			if (p == prev) break;
		}
		lstrcat(title, " [ ");
		lstrcat(title, p2);
		lstrcat(title, "]");
	}
	return true;
}


SEQUENCER *loadPmdFile(char *fn)
{
	PMDSEQ *p = new PMDSEQ;

	if (p == NULL) return NULL;

	if (setting.pcmdirnum) setpcmdir(setting.pcmdirlist);
	switch (music_load(fn))
	{
		case PMDWIN_OK:
		case WERNING_PPC_ALREADY_LOAD:
		case WERNING_P86_ALREADY_LOAD:
		case WERNING_PPS_ALREADY_LOAD:
		case WERNING_PPZ1_ALREADY_LOAD:
		case WERNING_PPZ2_ALREADY_LOAD:
			break;
		default:
		case ERR_OTHER:
			/* ? ? ? */
		case ERR_OPEN_MUSIC_FILE:
		case ERR_WRONG_MUSIC_FILE:
		case ERR_OPEN_PPC_FILE:
		case ERR_WRONG_PPC_FILE:
		case ERR_OPEN_P86_FILE:
		case ERR_WRONG_P86_FILE:
		case ERR_OPEN_PPS_FILE:
		case ERR_WRONG_PPS_FILE:
		case ERR_OPEN_PPZ1_FILE:
		case ERR_WRONG_PPZ1_FILE:
		case ERR_OPEN_PPZ2_FILE:
		case ERR_WRONG_PPZ2_FILE:
		case ERR_OUT_OF_MEMORY:
			delete p;
			return NULL;
	}
	p->loopc = 0;
	p->seq.work = p;
	p->seq.Play = Play;
	p->seq.Stop = Stop;
	p->seq.Mix = Mix;
	p->seq.Term = Term;
	p->seq.SetPosition = SetPosition;
	p->seq.GetPosition = GetPosition;
	p->seq.GetRate = GetRate;
	p->seq.GetPriority = GetPriority;
	p->seq.GetLoopCount = GetLoopCount;
	p->seq.IsPlaying = IsPlaying;
	return &p->seq;
}

int infoPmdBox(char *fn, HWND hwndParent)
{
	return 0;
}

int isOurPmdFile(char *fn)
{
	return 0;
}
